#include "cellmanagement.h"
#include"complex.h"
vector< vector<short int> > cellpos, relneighshift;
int numneighcells, totcells, nbs;
int powersofnc[10];

extern int N;  // # of ellipses
extern int d;  // spatial dimension
extern double alpha, sigma, skin; // ellipse aspect ratio, semiminor axis
extern vector< vector<double> > Rall, Rlast;
extern vector<double> rtrial;
extern double D, lo, hi; // cell side length, coordinates of bottom & top boundaries
extern int ncell;  // number of subcells along each side
extern int ninserted;
extern long long int ntrial;

// For SWAP
extern vector <double> Gi, gaps;

inline int mypowi(int a, int b);
/*Distance of Closest Approach of two arbitrary ellipses*/
extern double distance2d(double a1, double b1, double a2, double b2, double k1dotd, double k2dotd, double k1dotk2);
/* Principal cubic root of a complex number */
extern complex<double> c_cbrt(complex<double> x);


// Using cellpos and relneighshift to handle neigbor-mapping
// elements of cellpos are (0, 1, ..., ncell - 1) along each dimension
// elements of relneighshift are (-1, 0, 1) along each dimension
// Could try using smaller cells (alpha/2 rather than alpha) and setting nbs = 5; see below
void cell::setupsupercell() {
    int i, f, h;
    totcells = mypowi(ncell,d);
    subcell.resize(totcells);   boxpops.resize(totcells);
    numneighcells = mypowi(3,d);  
    int shift = 1;

// define cellpos and relneighpos
    cellpos.resize(totcells);
    for (i = 0; i < totcells; i++) cellpos[i].resize(d);
    relneighshift.resize(numneighcells);
    for (i = 0; i < numneighcells; i++) { relneighshift[i].resize(d); }
    for (i = 0; i < totcells; i++) {
        h = i;
        for (f = 0; f < d; f++) {
            cellpos[i][d-1-f] = h%ncell;
            h /= ncell;
        }
    }
    for (i = 0; i < numneighcells; i++) {
        h = i;
// relneighshift = min-image separation vector between cell centers
        for (f = 0; f < d; f++) {
            relneighshift[i][d-1-f] = h%3 - shift;
            h /= 3;
        }
    }
    
// powersofnc used to determine neighbor l of cell i in buildneighborlist()
    for (i = 0; i < d; i++) powersofnc[i] = mypowi(ncell,i);
    
// Set initial boxpops = 0
    for (i = 0; i < totcells; i++) {  boxpops[i] = 0; subcell[i].atomlist.clear();  }
    
    return;
}



// Place the ellipses in their subcells
void cell::populatecells() {
    int f, h, i, j, k, l, m, n, o, p, q, r, u, v;
    int to[d];    // used to place atoms in bins
    double delc[d], delc0[d], delsq;
    double rangerq; // for polydispersity-aware neighbor-list building
    double totneigh = 0;

// Reset subcell & atomic-neighbor lists
    for (i = 0; i < totcells; i++) {    subcell[i].atomlist.clear();  boxpops[i] = 0;  }

// Put atoms into their subcells
    for (q = 0; q < N; q++) {
        f = 0;
        for (v = 0; v < d; v++) {
            to[v] = floor(((Rall[q][v] - lo)/(D/ncell)));
// Cleanly exit program if an atom gets lost
            if (to[v] < 0 || to[v] > ncell-1) {
                cout << "Error on trial " << ntrial << ": Atom " << q << " escaped the simulation cell, current position is";
                for (i = 0; i < d-1; i++) cout << " " << Rall[q][i];
                exit(1);
            }
            f += to[v]*powersofnc[d-1-v];
        }
        subcell[f].atomlist.emplace_back(q);
        boxpops[f] += 1;
    }

// Update Rlast
    for (i = 0; i < N; i++) for (j = 0; j < d+2; j++) Rlast[i][j] = Rall[i][j];
    return;
}



// Check whether new trial position overlaps with particles in same and neighboring subcells
bool cell::doesitfit(int q, bool RSAflag) {
    int f, h, i, j, k, l, m, n, o, p, r, u, v;
    int to[d];    // used to place atoms in bins
    double delvec[d], del, delsq, deltatheta;
    int neighindex;
    int neighpos[numneighcells*d];
    bool fitflag = true;
// Ellipse-geometry variables
    double a, asq, b, bsq, maxdistsq, mindistsq;
// Quantities defined in Zheng and Palffy-Muhoray, PRE 75, 061709 (2007)
    double dhat[d], k1[d], k2[d];
    double dotprod0, dotprod1, dotprod2, theta1, theta2, d_cap;

// Determine which subcell "f" the trial particle should be placed in
    f = 0;
    for (v = 0; v < d; v++) {
        to[v] = floor(((rtrial[v]- lo)/(D/ncell)));
        f += to[v]*powersofnc[d-1-v];
    }
// Loop over the numneighcells neighbors of cell f, checking for overlaps of the trial particle
// position with all particles in these cells
    for (l = 0; l < numneighcells; l++) {
// Determine the current neighbor-subcell index
        neighindex = 0;
        for (v = 0; v < d; v++) {
            neighpos[l*d+v] = cellpos[f][v] + relneighshift[l][v];
            if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
            else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
        }
        for (v = 0; v < d; v++) neighindex += neighpos[l*d+v]*powersofnc[d-1-v];
// Loop over atoms in the current neighbor-subcell
        for (p = 0; p < boxpops[neighindex]; p++) {
            r = subcell[neighindex].atomlist[p];
            if (r == q) continue;  // don't check for self-overlap!
            delvec[0] = Rall[r][0] - rtrial[0];
            delvec[1] = Rall[r][1] - rtrial[1];
            delsq = 0;
            for (v = 0; v < 2; v++) {   // Apply PBCs to delvec
                if (delvec[v] < -D/2.0) delvec[v] += D;
                else if (delvec[v] > D/2.0) delvec[v] -= D;
                delsq += delvec[v]*delvec[v];
            }
// Set the overlap criteria for polydisperse systems
            b = (rtrial[3]+Rall[r][3])/2;       bsq = b*b;
            a = alpha*b;                        asq = a*a;
            maxdistsq = 4*asq;                  mindistsq = 4*bsq;
// Ellipses overlap if delsq < b^2, don't overlap if delsq > a^2
            if (delsq < mindistsq) { fitflag = false;  return fitflag; }
            else if (delsq > maxdistsq) continue;
// Don't need go any further for alpha = 1
            if (alpha == 1) continue;
            
// Now calculate d_cap using Xiaoyu Zheng's routine for calculating d_cap
// See http://www.math.kent.edu/~zheng/ellipse.html
            k1[0] = cos(rtrial[2]);     k1[1] = sin(rtrial[2]);
            k2[0] = cos(Rall[r][2]);    k2[1] = sin(Rall[r][2]);
            dotprod0 = k1[0]*k2[0] + k1[1]*k2[1];
            deltatheta = acos(dotprod0);
            del = sqrt(delsq);
            dhat[0] = delvec[0]/del;    dhat[1] = delvec[1]/del;
            dotprod1 = k1[0]*dhat[0] + k1[1]*dhat[1];
            dotprod2 = k2[0]*dhat[0] + k2[1]*dhat[1];
            theta1 = acos(dotprod1);    theta2 = acos(dotprod2);
// Old version for monodisperse systems
//            d_cap = distance2d(a,b,a,b,dotprod1,dotprod2,dotprod0);
// Polydisperse version
            d_cap = distance2d(alpha*rtrial[3],rtrial[3],alpha*Rall[r][3],Rall[r][3],dotprod1,dotprod2,dotprod0);
            if (del < d_cap) {   fitflag = false;    return fitflag;  }
        } // end of the p-loop
    } // end of the l-loop

// During the initial-state generation, if code gets to here, the trial particle fits: Add it to the atomlist of subcell f and append its coordinates to Rall
    if (RSAflag) {
        subcell[f].atomlist.emplace_back(q);
        boxpops[f] += 1;
        ninserted++;
        Rall.push_back(rtrial);
    }
    
    return fitflag;
}


// Find the minimum d - d_cap for all particles
double cell::findmingapdist(bool diamreductionflag) {
    int f, h, i, j, k, l, m, n, o, p, q, r, u, v;
    int to[d];    // used to place atoms in bins
    double delvec[d], del, delsq, deltatheta;
    int neighindex;
    int neighpos[numneighcells*d];
// Ellipse-geometry variables
    double a, asq, b, bsq, maxdistsq, mindistsq;
// Quantities defined in Zheng and Palffy-Muhoray, PRE 75, 061709 (2007)
    double dhat[d], k1[d], k2[d];
    double dotprod0, dotprod1, dotprod2, theta1, theta2, d_cap;
    double minGi = 1000;
    bool ovflag[N];
    
for (q = 0; q < N; q++) {
    Gi[q] = 1000;    gaps[q] = 1000;    ovflag[q] = false;
// Determine which subcell "f" the trial particle should be placed in
    f = 0;
    for (v = 0; v < d; v++) {
        to[v] = floor(((Rall[q][v]- lo)/(D/ncell)));
        f += to[v]*powersofnc[d-1-v];
    }
// Loop over the numneighcells neighbors of cell f, checking for overlaps of particle q
// with all particles in these cells
    for (l = 0; l < numneighcells; l++) {
// Determine the current neighbor-subcell index
        neighindex = 0;
        for (v = 0; v < d; v++) {
            neighpos[l*d+v] = cellpos[f][v] + relneighshift[l][v];
            if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
            else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
        }
        for (v = 0; v < d; v++) neighindex += neighpos[l*d+v]*powersofnc[d-1-v];
// Loop over atoms in the current neighbor-subcell
        for (p = 0; p < boxpops[neighindex]; p++) {
            r = subcell[neighindex].atomlist[p];
            if (r == q) continue;  // don't check for self-overlap!
            delvec[0] = Rall[r][0] - Rall[q][0];
            delvec[1] = Rall[r][1] - Rall[q][1];
            delsq = 0;
            for (v = 0; v < 2; v++) {   // Apply PBCs to delvec
                if (delvec[v] < -D/2.0) delvec[v] += D;
                else if (delvec[v] > D/2.0) delvec[v] -= D;
                delsq += delvec[v]*delvec[v];
            }
// Set the overlap criteria for polydisperse systems
            b = (Rall[q][3]+Rall[r][3])/2;       bsq = b*b;
            a = alpha*b;                        asq = a*a;
            maxdistsq = 4*asq;                  mindistsq = 4*bsq;
// Ellipses overlap if delsq < b^2, don't overlap if delsq > a^2
            
// Now calculate d_cap using Xiaoyu Zheng's routine for calculating d_cap
// See http://www.math.kent.edu/~zheng/ellipse.html
            k1[0] = cos(Rall[q][2]);     k1[1] = sin(Rall[q][2]);
            k2[0] = cos(Rall[r][2]);    k2[1] = sin(Rall[r][2]);
            dotprod0 = k1[0]*k2[0] + k1[1]*k2[1];
            deltatheta = acos(dotprod0);
            del = sqrt(delsq);
            dhat[0] = delvec[0]/del;    dhat[1] = delvec[1]/del;
            dotprod1 = k1[0]*dhat[0] + k1[1]*dhat[1];
            dotprod2 = k2[0]*dhat[0] + k2[1]*dhat[1];
            theta1 = acos(dotprod1);    theta2 = acos(dotprod2);
// Old version for monodisperse systems
//            d_cap = distance2d(a,b,a,b,dotprod1,dotprod2,dotprod0);
// Polydisperse version
            d_cap = distance2d(alpha*Rall[q][3],Rall[q][3],alpha*Rall[r][3],Rall[r][3],dotprod1,dotprod2,dotprod0);
// NEW ON 1/19/24: IF AN OVERLAP OCCURS, ADAPTIVELY REDUCE THE PARTICLE DIAMETERS UNTIL IT VANISHES
// NEW ON 8/19/24: DO THIS ONLY IF diamreductionflag = true
            if (del < d_cap && diamreductionflag) {
                cout << setprecision(10) << fixed;
                cout << "Warning: on trial " << ntrial << ": findminGi detected a particle overlap: i = " << q << " j = " << r << " r_ij = " << del << " d_cap = " << d_cap << endl;
                ovflag[q] = true;
                do {
                    Rall[q][3] *= 0.99999999;        Rall[r][3] *= 0.99999999;
                    d_cap = distance2d(alpha*Rall[q][3],Rall[q][3],alpha*Rall[r][3],Rall[r][3],dotprod1,dotprod2,dotprod0);
                } while (del < d_cap);
                cout << "Reduced particle diameters so that r_ij = " << del << " d_cap = " << d_cap << endl;
            }
// Gi[q] is the recommended growth factor for particle q
            if (((del-d_cap)*Rall[q][3]/(Rall[q][3]+Rall[r][3]))/(2*alpha) < Gi[q]) Gi[q] = ((del-d_cap)*Rall[q][3]/(Rall[q][3]+Rall[r][3]))/(2*alpha);
// gaps[q] is the gap (gtilde) for particle[q]
            if (del - d_cap < gaps[q]) gaps[q] = del - d_cap;
        } // end of the p-loop
    } // end of the l-loop
// minGi is the recommended growth for particles that DIDN'T encounter overlaps
    if (Gi[q] < minGi) minGi = Gi[q];
}
 
    return minGi;
}


// Find the minimum d - d_cap for all particles
void cell::getlocalmingapdist(int q, bool diamreductionflag, bool neighflag) {
    int f, h, i, j, k, l, m, n, o, p, r, u, v;
    int to[d];    // used to place atoms in bins
    double delvec[d], del, delsq, deltatheta;
    int neighindex;
    int neighpos[numneighcells*d];
// Ellipse-geometry variables
    double a, asq, b, bsq, maxdistsq, mindistsq;
// Quantities defined in Zheng and Palffy-Muhoray, PRE 75, 061709 (2007)
    double dhat[d], k1[d], k2[d];
    double dotprod0, dotprod1, dotprod2, theta1, theta2, d_cap;
    double minGi = 1000;
    bool ovflag;
    
    Gi[q] = 1000;    gaps[q] = 1000;   ovflag = false;
// Determine which subcell "f" the trial particle should be placed in
    f = 0;
    for (v = 0; v < d; v++) {
        to[v] = floor(((Rall[q][v]- lo)/(D/ncell)));
        f += to[v]*powersofnc[d-1-v];
    }
// Loop over the numneighcells neighbors of cell f, checking for overlaps of particle q
// with all particles in these cells
    for (l = 0; l < numneighcells; l++) {
// Determine the current neighbor-subcell index
        neighindex = 0;
        for (v = 0; v < d; v++) {
            neighpos[l*d+v] = cellpos[f][v] + relneighshift[l][v];
            if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
            else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
        }
        for (v = 0; v < d; v++) neighindex += neighpos[l*d+v]*powersofnc[d-1-v];
// Loop over atoms in the current neighbor-subcell
        for (p = 0; p < boxpops[neighindex]; p++) {
            r = subcell[neighindex].atomlist[p];
            if (r == q) continue;  // don't check for self-overlap!
            delvec[0] = Rall[r][0] - Rall[q][0];
            delvec[1] = Rall[r][1] - Rall[q][1];
            delsq = 0;
            for (v = 0; v < 2; v++) {   // Apply PBCs to delvec
                if (delvec[v] < -D/2.0) delvec[v] += D;
                else if (delvec[v] > D/2.0) delvec[v] -= D;
                delsq += delvec[v]*delvec[v];
            }
// Set the overlap criteria for polydisperse systems
            b = (Rall[q][3]+Rall[r][3])/2;       bsq = b*b;
            a = alpha*b;                        asq = a*a;
            maxdistsq = 4*asq;                  mindistsq = 4*bsq;
// Ellipses overlap if delsq < b^2, don't overlap if delsq > a^2
            
// Now calculate d_cap using Xiaoyu Zheng's routine for calculating d_cap
// See http://www.math.kent.edu/~zheng/ellipse.html
            k1[0] = cos(Rall[q][2]);     k1[1] = sin(Rall[q][2]);
            k2[0] = cos(Rall[r][2]);    k2[1] = sin(Rall[r][2]);
            dotprod0 = k1[0]*k2[0] + k1[1]*k2[1];
            deltatheta = acos(dotprod0);
            del = sqrt(delsq);
            dhat[0] = delvec[0]/del;    dhat[1] = delvec[1]/del;
            dotprod1 = k1[0]*dhat[0] + k1[1]*dhat[1];
            dotprod2 = k2[0]*dhat[0] + k2[1]*dhat[1];
            theta1 = acos(dotprod1);    theta2 = acos(dotprod2);
// Old version for monodisperse systems
//            d_cap = distance2d(a,b,a,b,dotprod1,dotprod2,dotprod0);
// Polydisperse version
            d_cap = distance2d(alpha*Rall[q][3],Rall[q][3],alpha*Rall[r][3],Rall[r][3],dotprod1,dotprod2,dotprod0);
// NEW ON 1/19/24: IF AN OVERLAP OCCURS, ADAPTIVELY REDUCE THE PARTICLE DIAMETERS UNTIL IT VANISHES
// Should include the diamreductionflag conditional here?
            if (del < d_cap) {
                cout << setprecision(10) << fixed;
                cout << "Warning: on trial " << ntrial << ": findminGi detected a particle overlap: i = " << q << " j = " << r << " r_ij = " << del << " d_cap = " << d_cap << endl;
                ovflag = true;
                do {
                    Rall[q][3] *= 0.99999999;        Rall[r][3] *= 0.99999999;
                    d_cap = distance2d(alpha*Rall[q][3],Rall[q][3],alpha*Rall[r][3],Rall[r][3],dotprod1,dotprod2,dotprod0);
                } while (del < d_cap);
                cout << "Reduced particle diameters so that r_ij = " << del << " d_cap = " << d_cap << endl;
            }
// Gi[q] is the recommended growth for particle q IF an overlap was generated
            if (((del-d_cap)*Rall[q][3]/(Rall[q][3]+Rall[r][3]))/(2*alpha) < Gi[q]) Gi[q] = ((del-d_cap)*Rall[q][3]/(Rall[q][3]+Rall[r][3]))/(2*alpha);
// gaps[q] is the gap (gtilde) for particle[q]
            if (del - d_cap < gaps[q]) gaps[q] = del - d_cap;
        } // end of the p-loop
    } // end of the l-loop

// Now do the same for all of particle q's neighbors
    if (neighflag) for (l = 0; l < numneighcells; l++) {
        // Determine the current neighbor-subcell index
        neighindex = 0;
        for (v = 0; v < d; v++) {
            neighpos[l*d+v] = cellpos[f][v] + relneighshift[l][v];
            if (neighpos[l*d+v] == -1) neighpos[l*d+v] = ncell - 1;
            else if (neighpos[l*d+v] == ncell) neighpos[l*d+v] = 0;
        }
        for (v = 0; v < d; v++) neighindex += neighpos[l*d+v]*powersofnc[d-1-v];
        // Loop over atoms in the current neighbor-subcell
        for (p = 0; p < boxpops[neighindex]; p++) {
            r = subcell[neighindex].atomlist[p];
            if (r != q) getlocalmingapdist(r,diamreductionflag,0); // note that now neighflag = false so this doesn't iterate beyond particle q's neighbors
        }
    }
    return;
}


// Raise an integer to an integer power
inline int mypowi(int a, int b) {
    int c, i;
    c = 1;
    for (i = 0; i < b; i++) c *= a;
    return c;
}

